#!/usr/bin/env python

# This file mainly exists to allow python setup.py test to work.
#
# You can test all the variations of tests by running:
#
#   ./manage.py test && python runtests.py && ./setup.py test && echo OK
#

import os, sys
os.environ['DJANGO_SETTINGS_MODULE'] = 'test_settings'

from django.core.management import call_command

def runtests():
    # use the call_command approach so that we are as similar to running
    # './manage.py test' as possible. Notably we need the South migrations to be
    # run.
    call_command('test', verbosity=2)
    sys.exit(0)

if __name__ == '__main__':
    runtests()



# The admin in deliberately left blank.
#
# As the data stored in this app will primarily come from a PopIt API we don't
# want to let people edit it locally, usless they really know what they are
# doing.
#
# In time perhaps this admin will be smart enough that it can tell what data is
# from a live popit API instance, and what is locally created, and allow you to
# edit accordingly.



from django.core.management.base import BaseCommand
from popit.models import ApiInstance

class Command(BaseCommand):
    help = 'Retrieve all documents from all PopIt API instances'

    def handle(self, *args, **options):
        for instance in ApiInstance.objects.all():
            instance.fetch_all_from_api()



import re

from django.db import models

from south.modelsinspector import add_introspection_rules


add_introspection_rules([], ["^popit\.fields\.ApiInstanceURLField"])

class ApiInstanceURLField(models.URLField):
    # Have a specific field for this because we might want to add some smarts at
    # some point, and we want to check it is actually a url to prevent odd errors
    # from happening when bad values are entered and then subsequently tried to be
    # used when fetching data. By default model validation only happens when
    # submitted through a form (which is silly, but there you go).
        
    description = "A url field that enforces validation"

    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        value = super(ApiInstanceURLField, self).to_python(value)

        # Want immediate validation on creation
        for validator in self.validators:
            validator(value)

        return value


add_introspection_rules([], ["^popit\.fields\.PopItURLField"])

class PopItURLField(models.URLField):
    # This field has some specific requirements - namely it is optional (locally
    # created docs won't have it) but if it has a value it should be unique. Should
    # also be a URL.
    #
    # This mean that it should store a NULL in the db when it is empty so that the
    # DB constraints act as expected (default Django behaviour would be to store an
    # empty string).
    #
    # NOTE - this relies on the database behaviour regarding UNIQUE constraints, and
    # NULL fields. If your DB does not allow several NULLs in a UNIQUE index then
    # this won't work. Will work in Postgres and SQLite3, might work in mySQL
    # depending on table type, probably won't work in SQLServer.
        
    description = "A url field that enforces validation and uniqueness if not empty"
    
    __metaclass__ = models.SubfieldBase
    
    def __init__(self, *args, **kwargs):

        # enforce some of the options
        kwargs['unique']  = True
        kwargs['blank']   = True
        kwargs['null']    = True
        kwargs['default'] = ''

        super(PopItURLField, self).__init__(*args, **kwargs)

    def to_python(self, value):
        value = super(PopItURLField, self).to_python(value)

        # convert None to empty string
        if value is None:
            value = ''

        # Want immediate validation on creation (if there is a value)
        if value != '':
            for validator in self.validators:
                validator(value)

        return value

    def get_db_prep_value(self, value, **kwargs):
        # return None if empty string - so that null is stored in db
        # see http://stackoverflow.com/a/5114226/5349 for partial source
        if value == "":
            return None
        else:
            return value



from django.test import TestCase
from django.db import IntegrityError
from django.core.exceptions import ValidationError

from popit.models import ApiInstance, Person

class PersonTest(TestCase):
    def test_instance_is_required(self):
        """
        Test that instance is required
        """
        # create instance
        instance = ApiInstance(url="http://foo.com/api")
        instance.save()
        self.assertTrue(instance)
        
        # create person
        person = Person(api_instance=instance, name="Bob")
        person.save()
        self.assertTrue(person)

        # try to create without instance
        self.assertRaises(IntegrityError, Person.objects.create, name="Bob") # instance missing
    

    def test_popit_url(self):
        """
        Test that the popit_url field is unique if it has a value and that it is
        possible to have several with no value.
        """
        
        # create instance
        instance = ApiInstance(url="http://foo.com/api")
        instance.save()
        self.assertTrue(instance)
        
        # create a person with no popit_url
        person_no_url_1 = Person(api_instance=instance, name="Bob")
        person_no_url_1.save()
        self.assertTrue(person_no_url_1)
        
        # create another
        person_no_url_2 = Person(api_instance=instance, name="Bob")
        person_no_url_2.save()
        self.assertTrue(person_no_url_2)
        
        # check that the empty popit_url is returned as empty string
        self.assertEqual(person_no_url_1.popit_url, '')
        
        # check it is empty string when fetched from db
        self.assertEqual( Person.objects.get(pk=person_no_url_1.id).popit_url, '')
        
        # url to use when testing
        popit_url = popit_url=instance.url + '/person/123'

        # create one with a popit_url
        person_with_url = Person(api_instance=instance, name="Bob", popit_url=popit_url)
        person_with_url.save()
        self.assertTrue(person_no_url_1)
        
        # try to create a duplicate one
        self.assertRaises(IntegrityError, Person.objects.create, api_instance=instance, name="Bob", popit_url=popit_url)

        # try to create one with a bad URL
        self.assertRaises(ValidationError, Person.objects.create, api_instance=instance, name="Bob", popit_url='This is not a URL')

        # try to create one with a URL that does not start with the ApiInstance URL
        # (implying that it is from a different instance)
        self.assertRaises(ValidationError, Person.objects.create, api_instance=instance, name="Bob", popit_url='http://other.com/api/person/123')
        


# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Person'
        db.create_table(u'popit_person', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('api_instance', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['popit.ApiInstance'])),
            ('popit_url', self.gf('popit.fields.PopItURLField')(default='', max_length=200, unique=True, null=True, blank=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=200)),
        ))
        db.send_create_signal(u'popit', ['Person'])


    def backwards(self, orm):
        # Deleting model 'Person'
        db.delete_table(u'popit_person')


    models = {
        u'popit.apiinstance': {
            'Meta': {'object_name': 'ApiInstance'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'url': ('popit.fields.ApiInstanceURLField', [], {'unique': 'True', 'max_length': '200'})
        },
        u'popit.person': {
            'Meta': {'object_name': 'Person'},
            'api_instance': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['popit.ApiInstance']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '200'}),
            'popit_url': ('popit.fields.PopItURLField', [], {'default': "''", 'max_length': '200', 'unique': 'True', 'null': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['popit']


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages


with open('requirements.txt') as f:
    required = f.read().splitlines()


setup(
    name='popit-django',
    version='0.0.1',
    description='Use PopIt stored data in your Django app',
    license='GNU Affero General Public License v3',
    author='mySociety',
    author_email='modules@mysociety.org',
    url='https://github.com/mysociety/popit-django',
    long_description=open('README.rst', 'r').read(),

    # Not convinced that this is correct - it skips the fixture which are .js files
    packages=find_packages(),

    requires=[],
    install_requires=required,
    tests_require=[],
    classifiers=[
        # choose from https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    test_suite='runtests.runtests',
)



"""Settings that need to be set in order to run the tests."""

import os

DEBUG = True
USE_TZ = True
SITE_ID = 1

SECRET_KEY = '...something secure here...'

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME":   "popit-django.sqlite3",
    }
}

ROOT_URLCONF = 'popit.tests.urls'

CURRENT_DIR = os.path.dirname(__file__)
STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(CURRENT_DIR, '../../static/')
STATICFILES_DIRS = (
    os.path.join(CURRENT_DIR, 'test_static'),
)

TEMPLATE_DIRS = (
    os.path.join(CURRENT_DIR, '../templates'),
)

INSTALLED_APPS = [
    'south',
    'popit',
]

# Testing related
TEST_POPIT_API_HOST_IP   = '127.0.0.1'
TEST_POPIT_API_PORT      = '3000'
TEST_POPIT_API_SUBDOMAIN = 'popit-django-test'

# create the url to use for testing the database.
# See http://xip.io/ for details on the domain used.
TEST_POPIT_API_URL = "http://%s.%s.xip.io:%s/api" % ( TEST_POPIT_API_SUBDOMAIN,
                                                      TEST_POPIT_API_HOST_IP,
                                                      TEST_POPIT_API_PORT )

# If you want to create a test entry this is useful:
# curl                                           \
#     -v                                         \
#     -H "Content-type: application/json"        \
#     -X POST                                    \
#     -d ' {"name": "Joe Bloggs"}'               \
#     http://popit-django-test.127.0.0.1.xip.io:3000/api/persons/



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'ApiInstance'
        db.create_table(u'popit_apiinstance', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('url', self.gf('popit.fields.ApiInstanceURLField')(unique=True, max_length=200)),
        ))
        db.send_create_signal(u'popit', ['ApiInstance'])


    def backwards(self, orm):
        # Deleting model 'ApiInstance'
        db.delete_table(u'popit_apiinstance')


    models = {
        u'popit.apiinstance': {
            'Meta': {'object_name': 'ApiInstance'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'url': ('popit.fields.ApiInstanceURLField', [], {'unique': 'True', 'max_length': '200'})
        }
    }

    complete_apps = ['popit']


import slumber

from django.db import models
from django.core.exceptions import ValidationError

from .fields import ApiInstanceURLField, PopItURLField

class ApiInstance(models.Model):
    """A specific PopIt API instance"""
    url = ApiInstanceURLField(
        unique    = True,
        help_text = "The full URL to the root of the api - eg 'http://example.com/api"
    )

    ### future fields might include:
    # name
    # last_checked - for incremental updates

    def api_client(self, collection_name):
        api = slumber.API(self.url).__getattr__(collection_name)
        return api

    def fetch_all_from_api(self):
        """
        Update all the local data from the API. This method actually delegates
        to the other models.
        """
        
        models = [Person]
        for model in models:
            model.fetch_all_from_api(instance=self)


class PopItDocument(models.Model):
    # The API instance is required
    api_instance = models.ForeignKey('ApiInstance')
    popit_url    = PopItURLField()

    name         = models.CharField(max_length=200)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Add a check that the popit_url starts with the api_instance.url - so
        that we can be sure that the url is coming from the correct api.
        """
        if self.popit_url:
            try:
                self.popit_url.index(self.api_instance.url)
            except ValueError:
                raise ValidationError("popit_url does not start with its api_instance.url")

        super(PopItDocument, self).save(*args, **kwargs)


    # This could possibly live on the manager and be called using Foo.objects.fetch_all_from_api(), perhaps ...
    @classmethod
    def fetch_all_from_api(cls, instance):
        """
        Get all the documents from the API and save them locally.
        """

        api_client = instance.api_client(cls.api_collection_name)

        # This is hacky, but I can't see a documented way to get to the url.
        # Liable to change if slumber changes their internals.
        collection_url = api_client._store['base_url']

        for doc in api_client.get()['result']:
            url = collection_url + '/' + doc['id']
            
            # Add the url to the doc
            doc['popit_url'] = url
            
            cls.update_from_api_results(instance=instance, doc=doc)

    @classmethod
    def update_from_api_results(cls, instance, doc):

        # load the object from the db, or create it
        try:
            obj = cls.objects.get(popit_url=doc['popit_url'])
        except cls.DoesNotExist:
            obj = cls(api_instance=instance, popit_url=doc['popit_url'])

        # extract the dict of settable fields and update the obj
        settable = cls.extract_settable(doc)
        [setattr(obj, key, value) for (key, value) in settable.items()]

        obj.save()
        
        # TODO create/update related records here
        # obj.update_related_from_api_result(doc)


    @classmethod
    def extract_settable(cls, doc):
        """
        You should override this method to extract the fields from the API
        result that should be set on the top level object. The response should
        be a dict of keys that are the field names, and values that are what
        should be set to them.

        This only applies to fields unique to the class - generic ones like
        popit_url are set elsewhere.
        """
        raise NotImplementedError("Override extract_settable in '%s'" % cls)


class Person(PopItDocument):
    """A Person from a PopIt API instance"""

    api_collection_name = 'persons'

    @classmethod
    def extract_settable(cls, doc):
        return {
            'name': doc['name']
        }




import slumber
import os
import subprocess

from django.conf import settings


def get_api_url():
    return settings.TEST_POPIT_API_URL
    
def get_api_client():
    return slumber.API(get_api_url())

def get_api_database_name():
    api = get_api_client()
    # https://github.com/dstufft/slumber/issues/28
    return api.__getattr__('').get()['info']['databaseName']

def delete_api_database():
    name = get_api_database_name()
    dev_null = open(os.devnull, 'w')
    subprocess.call(["mongo", name, '--eval', 'db.dropDatabase()'], stdout=dev_null)
    
def load_test_data(fixture_name='default'):
    """
    Use the mongofixtures CLI tool provided by the pow-mongodb-fixtures package
    used by popit-api to load some test data into db. Don't use the test fixture
    from popit-api though as we don't want changes to that to break our test
    suite.

        https://github.com/powmedia/pow-mongodb-fixtures#cli-usage

    """
    
    project_root = os.path.normpath(os.path.join(os.path.dirname(__file__), '../..'))
    
    # gather the args for the call
    mongofixtures_path = os.path.join( project_root, 'popit-api-for-testing/node_modules/.bin/mongofixtures' )
    database_name      = get_api_database_name()
    test_fixtures_path = os.path.join( project_root, 'popit/tests/fixtures/%s.js'%fixture_name )

    # Check that the fixture exists
    if not os.path.exists(test_fixtures_path):
        raise Exception("Could not find fixture for %s at %s" % (fixture_name, test_fixtures_path))

    # Hack to deal with bad handling of absolute paths in mongofixtures.
    # Fix: https://github.com/powmedia/pow-mongodb-fixtures/pull/14
    test_fixtures_path = os.path.relpath( test_fixtures_path )

    # Usage: mongofixtures db_name path/to/fixtures.js
    dev_null = open(os.devnull, 'w')
    exit_code = subprocess.call([mongofixtures_path, database_name, test_fixtures_path], stdout=dev_null)
    if exit_code:
        raise Exception("Error loading fixtures for '%s'" % fixture_name)



from django.test import TestCase
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.core.management import call_command

from popit.models import ApiInstance, Person
from popit.tests import instance_helpers

class ApiInstanceTest(TestCase):
    def test_url_constraints(self):
        """
        Test that url is required and unique
        """
        # create a good one
        instance = ApiInstance(url="http://foo.com/api")
        instance.save()
        self.assertTrue(instance)
        
        # test that the url is required and must be a url - even when creating entries
        # from code.
        self.assertRaises(ValidationError, ApiInstance, url="not a url")
        self.assertRaises(ValidationError, ApiInstance) # url missing

        # test that we can't create a duplicate
        def save_duplicate():
            ApiInstance(url=instance.url).save()
        self.assertRaises(IntegrityError, save_duplicate)
        

    def test_retrieve_all_from_instance(self):
        """
        Test that it is possible to retrieve several people from the instance.
        """
    
        # create the instance, delete contents and load test fixture
        instance_helpers.delete_api_database()
        instance_helpers.load_test_data()
        instance = ApiInstance.objects.create(url=instance_helpers.get_api_url())
        self.assertTrue(instance)
    
        # Tell the instance to sync data
        instance.fetch_all_from_api()
    
        # check that the persons expected are loaded
        person = Person.objects.get(name='Joe Bloggs')
        self.assertTrue(person)

        # update the api to use a different fixture and check that the update is
        # applied
        instance_helpers.delete_api_database()
        instance_helpers.load_test_data('rename_joe_bloggs')
        instance.fetch_all_from_api()
        renamed = Person.objects.get(pk=person.id)
        self.assertEqual(renamed.name, 'Josh Blaggs')


    def test_retrieve_all_management_command(self):
        
        # create the instance, delete contents and load test fixture
        instance_helpers.delete_api_database()
        instance_helpers.load_test_data()
        ApiInstance.objects.create(url=instance_helpers.get_api_url())
        
        # call the management command to retrieve all
        call_command('popit_retrieve_all')
        
        # check that the persons expected are loaded
        person = Person.objects.get(name='Joe Bloggs')
        self.assertTrue(person)
        